@@ -157,7 +157,7 @@ vers="0.12"
 os.environ['LC_ALL'] = "C"
 
 # Argument parsing
-opts = "L:DnvVhd:r:"
+opts = "L:DnvVhd:r:l:"
 longopts = ["no-default-lib", "dry-run", "verbose", "version", "help",
             "dest-dir=", "ldlib=", "libc-extras-dir=", "root="]
 
@@ -195,6 +195,8 @@ for opt, arg in optlist:
         libc_extras_dir = arg
     elif opt in ("-r", "--root"):
         root = arg
+    elif opt in ("-l",):
+        pass
     elif opt in ("--help", "-h"):
 	usage(0)
         sys.exit(0)
