@@ -18,53 +18,16 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-#include "elf.hpp"
-#include "endian.hpp"
+#include "elf_data.hpp"
 
 #include <stdexcept>
 
 #include <fcntl.h>
-#include <elf.h>
 #include <sys/mman.h>
 #include <sys/stat.h>
 
 using namespace Elf;
 
-namespace
-{
-  template <typename _class>
-    struct _elfdef
-    { };
-
-  template <>
-    struct _elfdef<file_class_32>
-    {
-      typedef Elf32_Dyn Dyn;
-      typedef Elf32_Ehdr Ehdr;
-      typedef Elf32_Phdr Phdr;
-      typedef Elf32_Shdr Shdr;
-      typedef Elf32_Sym Sym;
-      static inline uint8_t st_bind (uint8_t st_info) throw () { return ELF32_ST_BIND (st_info); }
-      static inline uint8_t st_type (uint8_t st_info) throw () { return ELF32_ST_TYPE (st_info); }
-    };
-
-  template <>
-    struct _elfdef<file_class_64>
-    {
-      typedef Elf64_Dyn Dyn;
-      typedef Elf64_Ehdr Ehdr;
-      typedef Elf64_Phdr Phdr;
-      typedef Elf64_Shdr Shdr;
-      typedef Elf64_Sym Sym;
-      static inline uint8_t st_bind (uint8_t st_info) throw () { return ELF64_ST_BIND (st_info); }
-      static inline uint8_t st_type (uint8_t st_info) throw () { return ELF64_ST_TYPE (st_info); }
-    };
-}
-
-file::file (const char *filename, void *mem, size_t len) throw (std::bad_alloc)
-: filename (std::string (filename)), mem (mem), len (len)
-{ }
-
 file::~file () throw ()
 {
   ::munmap (mem, len);
@@ -93,9 +56,9 @@ file *file::open (const char *filename) throw (std::bad_alloc, std::runtime_erro
     switch (buf[EI_CLASS])
     {
       case ELFCLASS32:
-        return open_class<file_class_32> (filename, buf, mem, len);
+        return open_class<file_class_32> (mem, len);
       case ELFCLASS64:
-        return open_class<file_class_64> (filename, buf, mem, len);
+        return open_class<file_class_64> (mem, len);
       default:
         throw std::runtime_error ("Invalid file class");
     }
@@ -108,35 +71,24 @@ file *file::open (const char *filename) throw (std::bad_alloc, std::runtime_erro
 }
 
 template<typename _class>
-file *file::open_class (const char *filename, const uint8_t *buf, void * mem, size_t len) throw (std::bad_alloc, std::runtime_error)
+file *file::open_class (void * mem, size_t len) throw (std::bad_alloc, std::runtime_error)
 {
+  const uint8_t *buf = static_cast <uint8_t *> (mem);
+
   switch (buf[EI_DATA])
   {
     case ELFDATA2LSB:
-      return new file_data<_class, file_data_2LSB> (filename, mem, len);
+      return new file_data<_class, file_data_2LSB> (mem, len);
     case ELFDATA2MSB:
-      return new file_data<_class, file_data_2MSB> (filename, mem, len);
+      return new file_data<_class, file_data_2MSB> (mem, len);
     default:
       throw std::runtime_error ("Invalid file data");
   }
 }
 
 template <typename _class, typename _data>
-file_data<_class, _data>::file_data (const char *filename) throw (std::bad_alloc, std::runtime_error)
-: file (filename)
-{ 
-  construct ();
-}
-
-template <typename _class, typename _data>
-file_data<_class, _data>::file_data (const char *filename, void *mem, size_t len) throw (std::bad_alloc, std::runtime_error)
-: file (filename, mem, len)
-{
-  construct ();
-}
-
-template <typename _class, typename _data>
-void file_data<_class, _data>::construct () throw (std::bad_alloc, std::runtime_error)
+file_data<_class, _data>::file_data (void *mem, size_t len) throw (std::bad_alloc, std::runtime_error)
+: file (mem, len)
 {
   uint8_t *buf = static_cast <uint8_t *> (this->mem);
   if (buf[EI_CLASS] != _class::id)
@@ -167,11 +119,23 @@ void file_data<_class, _data>::construct () throw (std::bad_alloc, std::runtime_
     section *temp;
     switch (convert<_data, typeof (shdrs[i].sh_type)> () (shdrs[i].sh_type))
     {
+      case section_type_STRTAB::id:
+        temp = new section_real<_class, _data, section_type_STRTAB> (&shdrs[i], this->mem);
+        break;
       case section_type_DYNAMIC::id:
-        temp = new section_real<_class, _data, section_type_DYNAMIC> (&shdrs[i], this->mem);
+        temp = section_DYNAMIC = new section_real<_class, _data, section_type_DYNAMIC> (&shdrs[i], this->mem);
         break;
       case section_type_DYNSYM::id:
-        temp = new section_real<_class, _data, section_type_DYNSYM> (&shdrs[i], this->mem);
+        temp = section_DYNSYM = new section_real<_class, _data, section_type_DYNSYM> (&shdrs[i], this->mem);
+        break;
+      case section_type_GNU_VERDEF::id:
+        temp = section_GNU_VERDEF = new section_real<_class, _data, section_type_GNU_VERDEF> (&shdrs[i], this->mem);
+        break;
+      case section_type_GNU_VERNEED::id:
+        temp = section_GNU_VERNEED = new section_real<_class, _data, section_type_GNU_VERNEED> (&shdrs[i], this->mem);
+        break;
+      case section_type_GNU_VERSYM::id:
+        temp = section_GNU_VERSYM = new section_real<_class, _data, section_type_GNU_VERSYM> (&shdrs[i], this->mem);
         break;
       default:
         temp = new section_real<_class, _data, section_type_UNDEFINED> (&shdrs[i], this->mem);
@@ -191,7 +155,7 @@ void file_data<_class, _data>::construct () throw (std::bad_alloc, std::runtime_
     switch (convert<_data, typeof (phdrs[i].p_type)> () (phdrs[i].p_type))
     {
       case segment_type_INTERP::id:
-        temp = new segment_real<_class, _data, segment_type_INTERP> (&phdrs[i], this->mem);
+        temp = segment_INTERP = new segment_real<_class, _data, segment_type_INTERP> (&phdrs[i], this->mem);
         break;
       default:
         temp = new segment_real<_class, _data, segment_type_UNDEFINED> (&phdrs[i], this->mem);
@@ -200,14 +164,12 @@ void file_data<_class, _data>::construct () throw (std::bad_alloc, std::runtime_
     this->segments[i] = temp;
   }
 
-  for (unsigned int i = 0; i < this->shnum; i++)
-    this->sections[i]->update_string_table (this);
-}
-
-void section::update_string_table (file *file) throw (std::bad_alloc)
-{
-  const section *section = file->get_section (file->get_shstrndx ());
-  this->name_string = std::string (static_cast <const char *> (section->_mem ()) + this->name);
+  for (std::vector<section *>::iterator it = sections.begin(); it != sections.end(); ++it)
+  {
+    section_data<_class, _data> &section =
+      dynamic_cast <section_data<_class, _data> &> (**it);
+    section.update(*this);
+  }
 }
 
 template <typename _class, typename _data>
@@ -223,6 +185,14 @@ section_data<_class, _data>::section_data (void *header, void *mem) throw ()
   this->mem = static_cast <void *> (static_cast <char *> (mem) + this->offset);
 }
 
+template <typename _class, typename _data>
+void section_data<_class, _data>::update(const file &file) throw (std::bad_alloc)
+{
+  const section_type<section_type_STRTAB> &section =
+    dynamic_cast <const section_type<section_type_STRTAB> &> (file.get_section(file.get_shstrndx()));
+  name_string = section.get_string(name);
+}
+
 section_type<section_type_DYNAMIC>::~section_type () throw ()
 {
   for (std::vector<dynamic *>::iterator it = dynamics.begin (); it != dynamics.end (); ++it)
@@ -245,11 +215,20 @@ section_real<_class, _data, section_type_DYNAMIC>::section_real (void *header, v
     this->dynamics.push_back (new dynamic_data<_class, _data> (&dyns[i]));
 }
 
-void section_type<section_type_DYNAMIC>::update_string_table (file *file) throw (std::bad_alloc)
+template <typename _class, typename _data>
+void section_real<_class, _data, section_type_DYNAMIC>::update(const file &file) throw (std::bad_alloc)
 {
-  section::update_string_table (file);
-  for (unsigned int i = 0; i < dynamics.size (); i++)
-    this->dynamics[i]->update_string_table (file, link);
+  section_data<_class, _data>::update(file);
+
+  const section_type<section_type_STRTAB> &section =
+    dynamic_cast <const section_type<section_type_STRTAB> &> (file.get_section(link));
+
+  for (std::vector<dynamic *>::iterator it = dynamics.begin(); it != dynamics.end(); ++it)
+  {
+    dynamic_data<_class, _data> &dynamic =
+      dynamic_cast <dynamic_data<_class, _data> &> (**it);
+    dynamic.update_string(section);
+  }
 }
 
 section_type<section_type_DYNSYM>::~section_type () throw ()
@@ -258,13 +237,6 @@ section_type<section_type_DYNSYM>::~section_type () throw ()
     delete *it;
 }
 
-void section_type<section_type_DYNSYM>::update_string_table (file *file) throw (std::bad_alloc)
-{
-  section::update_string_table (file);
-  for (unsigned int i = 0; i < symbols.size (); i++)
-    this->symbols[i]->update_string_table (file, link);
-}
-
 template <typename _class, typename _data>
 section_real<_class, _data, section_type_DYNSYM>::section_real (void *header, void *mem) throw (std::bad_alloc)
 : section_data<_class, _data> (header, mem)
@@ -281,6 +253,132 @@ section_real<_class, _data, section_type_DYNSYM>::section_real (void *header, vo
     this->symbols.push_back (new symbol_data<_class, _data> (&syms[i]));
 }
 
+template <typename _class, typename _data>
+void section_real<_class, _data, section_type_DYNSYM>::update(const file &file) throw (std::bad_alloc)
+{
+  section_data<_class, _data>::update (file);
+
+  const section_type<section_type_STRTAB> &section =
+    dynamic_cast <const section_type<section_type_STRTAB> &> (file.get_section(link));
+
+  for (unsigned int i = 0; i < symbols.size(); i++)
+  {
+    symbol_data<_class, _data> &symbol =
+      dynamic_cast <symbol_data<_class, _data> &> (*symbols[i]);
+    symbol.update_string(section);
+    symbol.update_version(file, i);
+  }
+}
+
+const version_definition *section_type<section_type_GNU_VERDEF>::get_version_definition(uint16_t index) const throw ()
+{
+  for (std::vector<version_definition *>::const_iterator it = verdefs.begin(); it != verdefs.end(); ++it)
+    if ((*it)->get_ndx() == index)
+      return *it;
+  return NULL;
+}
+
+template <typename _class, typename _data>
+section_real<_class, _data, section_type_GNU_VERDEF>::section_real (void *header, void *mem) throw (std::bad_alloc)
+: section_data<_class, _data> (header, mem)
+{
+  if (this->type != SHT_GNU_verdef)
+    throw std::logic_error ("Wrong section type");
+
+  typedef typename _elfdef<_class>::Verdef Verdef;
+  char *act = static_cast<char *> (this->mem);
+  uint32_t next = 0;
+
+  // TODO: Use DT_VERDEFNUM!
+  do
+  {
+    Verdef *verdef = reinterpret_cast<Verdef *> (act);
+    verdefs.push_back(new version_definition_data<_class, _data> (verdef));
+    next = convert<_data, typeof (verdef->vd_next)> () (verdef->vd_next);
+    act += next;
+  }
+  while (next);
+}
+
+template <typename _class, typename _data>
+void section_real<_class, _data, section_type_GNU_VERDEF>::update(const file &file) throw (std::bad_alloc)
+{
+  section_data<_class, _data>::update(file);
+
+  const section_type<section_type_STRTAB> &section =
+    dynamic_cast <const section_type<section_type_STRTAB> &> (file.get_section (link));
+
+  for (std::vector<version_definition *>::iterator it = verdefs.begin(); it != verdefs.end(); ++it)
+  {
+    version_definition_data<_class, _data> &verdef =
+      dynamic_cast<version_definition_data<_class, _data> &> (**it);
+    verdef.update_string(section);
+  }
+}
+
+const version_requirement_entry *section_type<section_type_GNU_VERNEED>::get_version_requirement_entry(uint16_t index) const throw ()
+{
+  for (std::vector<version_requirement *>::const_iterator it = verneeds.begin(); it != verneeds.end(); ++it)
+    for (std::vector<version_requirement_entry *>::const_iterator it1 = (*it)->get_entries().begin(); it1 != (*it)->get_entries().end(); ++it1)
+      if ((*it1)->get_other() == index)
+        return *it1;
+  return NULL;
+}
+
+template <typename _class, typename _data>
+section_real<_class, _data, section_type_GNU_VERNEED>::section_real (void *header, void *mem) throw (std::bad_alloc)
+: section_data<_class, _data> (header, mem)
+{
+  if (this->type != SHT_GNU_verneed)
+    throw std::logic_error ("Wrong section type");
+
+  typedef typename _elfdef<_class>::Verneed Verneed;
+  char *act = static_cast<char *> (this->mem);
+  uint32_t next = 0;
+
+  // TODO: Use DT_VERNEEDNUM!
+  do
+  {
+    Verneed *verneed = reinterpret_cast<Verneed *> (act);
+    verneeds.push_back(new version_requirement_data<_class, _data> (verneed));
+    next = convert<_data, typeof (verneed->vn_next)> () (verneed->vn_next);
+    act += next;
+  }
+  while (next);
+}
+
+template <typename _class, typename _data>
+void section_real<_class, _data, section_type_GNU_VERNEED>::update(const file &file) throw (std::bad_alloc)
+{
+  section_data<_class, _data>::update(file);
+
+  const section_type<section_type_STRTAB> &section =
+    dynamic_cast <const section_type<section_type_STRTAB> &> (file.get_section(link));
+
+  for (std::vector<version_requirement *>::iterator it = verneeds.begin (); it != verneeds.end (); ++it)
+  {
+    version_requirement_data<_class, _data> &verneed =
+      dynamic_cast<version_requirement_data<_class, _data> &> (**it);
+    verneed.update_string(section);
+  }
+}
+
+template <typename _class, typename _data>
+section_real<_class, _data, section_type_GNU_VERSYM>::section_real (void *header, void *mem) throw (std::bad_alloc)
+: section_data<_class, _data> (header, mem)
+{
+  if (this->type != SHT_GNU_versym)
+    throw std::logic_error ("Wrong section type");
+  typedef typename _elfdef<_class>::Versym Versym;
+  Versym *versyms = static_cast <Versym *> (this->mem);
+  unsigned int max = this->size / sizeof (Versym);
+
+  this->versyms.reserve (max);
+
+  for (unsigned int i = 0; i < max; i++)
+    this->versyms.push_back (versyms[i]);
+}
+
 template <typename _class, typename _data>
 segment_data<_class, _data>::segment_data (void *header, void *mem) throw ()
 {
@@ -325,11 +423,32 @@ dynamic_data<_class, _data>::dynamic_data (void *mem) throw ()
 }
 
 template <typename _class, typename _data>
-void dynamic_data<_class, _data>::update_string_table (file *file, uint16_t s) throw (std::bad_alloc)
+void dynamic_data<_class, _data>::update_string(const section_type<section_type_STRTAB> &section) throw (std::bad_alloc)
 {
-  const section *section = file->get_section (s);
-  if (this->is_string)
-    this->val_string = std::string (static_cast <const char *> (section->_mem ()) + this->val);
+  if (is_string)
+    val_string = section.get_string(val);
+}
+
+std::string symbol::get_version () const throw (std::bad_alloc)
+{
+  if (verneed)
+    return verneed->get_name();
+  else if (verdef)
+    return verdef->get_names()[0];
+
+  return "Base";
+}
+
+std::string symbol::get_name_version () const throw (std::bad_alloc)
+{
+  std::string ver;
+
+  if (verneed)
+    ver = '@' + verneed->get_name();
+  else if (verdef)
+    ver = '@' + verdef->get_names()[0];
+
+  return name_string + ver;
 }
 
 template <typename _class, typename _data>
@@ -347,9 +466,104 @@ symbol_data<_class, _data>::symbol_data (void *mem) throw ()
 }
 
 template <typename _class, typename _data>
-void symbol_data<_class, _data>::update_string_table (file *file, uint16_t s) throw (std::bad_alloc)
+void symbol_data<_class, _data>::update_string(const section_type<section_type_STRTAB> &section) throw (std::bad_alloc)
 {
-  const section *section = file->get_section (s);
-  this->name_string = std::string (static_cast <const char *> (section->_mem ()) + this->name);
+  name_string = section.get_string(name);
+}
+
+template <typename _class, typename _data>
+void symbol_data<_class, _data>::update_version(const file &file, uint16_t index) throw (std::bad_alloc)
+{
+  if (!file.get_section_GNU_VERSYM())
+    return;
+
+  versym = file.get_section_GNU_VERSYM()->get_versyms().at(index);
+
+  if (versym == 0 || versym == 1)
+    return;
+
+  if (file.get_section_GNU_VERNEED())
+  {
+    verneed = file.get_section_GNU_VERNEED()->get_version_requirement_entry(versym);
+    if (verneed)
+      return;
+  }
+
+  if (file.get_section_GNU_VERDEF())
+  {
+    verdef = file.get_section_GNU_VERDEF()->get_version_definition(versym & 0x7fff);
+    if (verdef)
+      return;
+  }
+
+  throw std::runtime_error("Invalid version");
+}
+
+template <typename _class, typename _data>
+version_definition_data<_class, _data>::version_definition_data (Verdef *verdef) throw ()
+{
+  this->ndx    = convert<_data, typeof (verdef->vd_ndx)> () (verdef->vd_ndx);
+  uint16_t cnt = convert<_data, typeof (verdef->vd_cnt)> () (verdef->vd_cnt);
+  uint32_t aux = convert<_data, typeof (verdef->vd_aux)> () (verdef->vd_aux);
+
+  char *act = reinterpret_cast<char *> (verdef) + aux;
+
+  for (int i = 0; i < cnt; i++)
+  {
+    Verdaux *verdaux = reinterpret_cast<Verdaux *> (act);
+    uint32_t name = convert<_data, typeof (verdaux->vda_name)> () (verdaux->vda_name);
+    uint32_t next = convert<_data, typeof (verdaux->vda_next)> () (verdaux->vda_next);
+    names.push_back(name);
+    act += next;
+  }
+}
+
+template <typename _class, typename _data>
+void version_definition_data<_class, _data>::update_string(const section_type<section_type_STRTAB> &section) throw (std::bad_alloc)
+{
+  for (std::vector<uint32_t>::iterator it = names.begin(); it != names.end(); ++it)
+    names_string.push_back(section.get_string(*it));
+}
+
+template <typename _class, typename _data>
+version_requirement_data<_class, _data>::version_requirement_data (Verneed *verneed) throw ()
+{
+  uint16_t cnt = convert<_data, typeof (verneed->vn_cnt)> () (verneed->vn_cnt);
+  uint32_t aux = convert<_data, typeof (verneed->vn_aux)> () (verneed->vn_aux);
+
+  char *act = reinterpret_cast<char *> (verneed) + aux;
+
+  for (int i = 0; i < cnt; i++)
+  {
+    Vernaux *vernaux = reinterpret_cast<Vernaux *> (act);
+    entries.push_back(new version_requirement_entry_data<_class, _data> (vernaux));
+    uint32_t next = convert<_data, typeof (vernaux->vna_next)> () (vernaux->vna_next);
+    act += next;
+  }
+}
+
+template <typename _class, typename _data>
+void version_requirement_data<_class, _data>::update_string(const section_type<section_type_STRTAB> &section) throw (std::bad_alloc)
+{
+  for (std::vector<version_requirement_entry *>::iterator it = entries.begin(); it != entries.end(); ++it)
+  {
+    version_requirement_entry_data<_class, _data> &vernaux =
+      dynamic_cast<version_requirement_entry_data<_class, _data> &> (**it);
+    vernaux.update_string(section);
+  }
+}
+
+template <typename _class, typename _data>
+version_requirement_entry_data<_class, _data>::version_requirement_entry_data (Vernaux *vna) throw ()
+{
+  flags = convert<_data, typeof (vna->vna_flags)> () (vna->vna_flags);
+  other = convert<_data, typeof (vna->vna_other)> () (vna->vna_other);
+  name  = convert<_data, typeof (vna->vna_name)> ()  (vna->vna_name);
+}
+
+template <typename _class, typename _data>
+void version_requirement_entry_data<_class, _data>::update_string(const section_type<section_type_STRTAB> &section) throw (std::bad_alloc)
+{
+  name_string = section.get_string(name);
 }
 
