@@ -21,7 +21,7 @@
 #ifndef ELF_HPP
 #define ELF_HPP
 
-#include "endian.hpp"
+#include "elf_defs.hpp"
 
 #include <stdexcept>
 #include <string>
@@ -31,18 +31,10 @@
 
 namespace Elf
 {
-  class file_class_32 { public: static const uint8_t id = 1; };
-  class file_class_64 { public: static const uint8_t id = 2; };
-  class file_data_2LSB { public: static const uint8_t id = 1; };
-  class file_data_2MSB { public: static const uint8_t id = 2; };
-  class section_type_UNDEFINED { };
-  class section_type_DYNAMIC { public: static const uint8_t id = 6; };
-  class section_type_DYNSYM { public: static const uint8_t id = 11; };
-  class segment_type_UNDEFINED { };
-  class segment_type_INTERP { public: static const uint8_t id = 3; };
-
   class section;
+  template <typename _type> class section_type;
   class segment;
+  template <typename _type> class segment_type;
 
   class file
   {
@@ -57,18 +49,23 @@ namespace Elf
       const uint16_t get_shstrndx () const throw () { return shstrndx; }
 
       const std::vector <section *> get_sections () const throw () { return sections; };
-      const section *get_section (unsigned int i) const throw (std::out_of_range) { return sections.at(i); };
-      const std::vector <segment *> get_segments () const throw () { return segments; };
+      const section &get_section (unsigned int i) const throw (std::out_of_range) { return *sections.at(i); };
+      const section_type<section_type_DYNAMIC> *get_section_DYNAMIC () const throw () { return section_DYNAMIC; };
+      const section_type<section_type_DYNSYM> *get_section_DYNSYM () const throw () { return section_DYNSYM; };
+      const section_type<section_type_GNU_VERDEF> *get_section_GNU_VERDEF () const throw () { return section_GNU_VERDEF; };
+      const section_type<section_type_GNU_VERNEED> *get_section_GNU_VERNEED () const throw () { return section_GNU_VERNEED; };
+      const section_type<section_type_GNU_VERSYM> *get_section_GNU_VERSYM () const throw () { return section_GNU_VERSYM; };
 
-      const void *const _mem () const throw () { return mem; }
+      const std::vector <segment *> get_segments () const throw () { return segments; };
+      const segment_type<segment_type_INTERP> *get_segment_INTERP () const throw () { return segment_INTERP; };
 
       static file *open (const char *filename) throw (std::bad_alloc, std::runtime_error);
 
     protected:
-      file (const char *, void *, size_t len) throw (std::bad_alloc);
+      file (void *mem, size_t len) throw (std::bad_alloc) : mem (mem), len (len) { }
 
       template<typename _class>
-        static file *open_class (const char *, const uint8_t *, void *, size_t) throw (std::bad_alloc, std::runtime_error);
+        static file *open_class (void *, size_t) throw (std::bad_alloc, std::runtime_error);
 
       uint16_t type;
       uint16_t machine;
@@ -82,27 +79,19 @@ namespace Elf
       uint16_t shstrndx;
 
       std::vector <section *> sections;
+      section_type<section_type_DYNAMIC> *section_DYNAMIC;
+      section_type<section_type_DYNSYM> *section_DYNSYM;
+      section_type<section_type_GNU_VERDEF> *section_GNU_VERDEF;
+      section_type<section_type_GNU_VERNEED> *section_GNU_VERNEED;
+      section_type<section_type_GNU_VERSYM> *section_GNU_VERSYM;
+
       std::vector <segment *> segments;
+      segment_type<segment_type_INTERP> *segment_INTERP;
 
-      const std::string filename;
       void *mem;
       size_t len;
   };
 
-  template <typename _class, typename _data>
-    class file_data : public file
-    {
-      public:
-        file_data (const char *) throw (std::bad_alloc, std::runtime_error);
-        file_data (const char *, void *, size_t len) throw (std::bad_alloc, std::runtime_error);
-
-        const uint8_t get_class () const throw () { return _class::id; }
-        const uint8_t get_data () const throw () { return _data::id; }
-
-      private:
-        void construct () throw (std::bad_alloc, std::runtime_error);
-    };
-
   class section
   {
     public:
@@ -112,10 +101,6 @@ namespace Elf
       uint64_t get_size () const throw () { return size; }
       const std::string &get_name_string () const throw () { return name_string; }
 
-      const void *const _mem () const throw () { return mem; }
-
-      virtual void update_string_table (file *) throw (std::bad_alloc);
-
     protected:
       uint32_t name;
       uint32_t type;
@@ -128,13 +113,6 @@ namespace Elf
       void *mem;
   };
 
-  template <typename _class, typename _data>
-    class section_data : public virtual section
-    {
-      public:
-        section_data (void *, void *) throw ();
-    };
-
   template <typename _type>
     class section_type : public virtual section
     {
@@ -142,6 +120,19 @@ namespace Elf
 
   class dynamic;
   class symbol;
+  class version_definition;
+  class version_requirement;
+  class version_requirement_entry;
+
+  template <>
+    class section_type<section_type_STRTAB> : public virtual section
+    {
+      public:
+        std::string get_string (uint32_t offset) const throw (std::bad_alloc)
+        {
+          return std::string (static_cast<const char *> (mem) + offset);
+        }
+    };
 
   template <>
     class section_type<section_type_DYNAMIC> : public virtual section
@@ -149,9 +140,7 @@ namespace Elf
       public:
         ~section_type () throw ();
 
-        const std::vector<dynamic *> &get_dynamics () throw () { return dynamics; }
-
-        void update_string_table (file *) throw (std::bad_alloc);
+        const std::vector<dynamic *> &get_dynamics () const throw () { return dynamics; }
 
       protected:
         std::vector<dynamic *> dynamics;
@@ -163,40 +152,48 @@ namespace Elf
       public:
         ~section_type () throw ();
 
-        const std::vector<symbol *> &get_symbols () throw () { return symbols; }
-
-        void update_string_table (file *) throw (std::bad_alloc);
+        const std::vector<symbol *> &get_symbols () const throw () { return symbols; }
 
       protected:
         std::vector<symbol *> symbols;
     };
 
-  template <typename _class, typename _data, typename _type>
-    class section_real : public section_data<_class, _data>, public section_type<_type>
+  template <>
+    class section_type<section_type_GNU_VERDEF> : public virtual section
     {
       public:
-        section_real (void *, void *) throw ();
+        ~section_type () throw () { }
+
+        const version_definition *get_version_definition(uint16_t index) const throw ();
+        const std::vector<version_definition *> &get_version_definitions () throw () { return verdefs; }
+
+      protected:
+        std::vector<version_definition *> verdefs;
     };
 
-  template <typename _class, typename _data>
-    class section_real<_class, _data, section_type_UNDEFINED> : public section_data<_class, _data>, public section_type<section_type_UNDEFINED>
+  template <>
+    class section_type<section_type_GNU_VERNEED> : public virtual section
     {
       public:
-        section_real (void *a, void *b) throw () : section_data<_class, _data> (a, b) { }
+        ~section_type () throw () { }
+
+        const version_requirement_entry *get_version_requirement_entry(uint16_t index) const throw ();
+        const std::vector<version_requirement *> &get_version_requirements () throw () { return verneeds; }
+
+      protected:
+        std::vector<version_requirement *> verneeds;
     };
 
-  template <typename _class, typename _data>
-    class section_real<_class, _data, section_type_DYNAMIC> : public section_data<_class, _data>, public section_type<section_type_DYNAMIC>
+  template <>
+    class section_type<section_type_GNU_VERSYM> : public virtual section
     {
       public:
-        section_real (void *, void *) throw (std::bad_alloc);
-    };
+        ~section_type () throw () { }
 
-  template <typename _class, typename _data>
-    class section_real<_class, _data, section_type_DYNSYM> : public section_data<_class, _data>, public section_type<section_type_DYNSYM>
-    {
-      public:
-        section_real (void *, void *) throw (std::bad_alloc);
+        const std::vector<uint16_t> &get_versyms () const throw () { return versyms; }
+
+      protected:
+        std::vector<uint16_t> versyms;
     };
 
   class segment
@@ -208,8 +205,6 @@ namespace Elf
       uint32_t get_flags () const throw () { return flags; }
       uint64_t get_filesz () const throw () { return filesz; }
 
-      const void *const _mem () const throw () { return mem; }
-
     protected:
       uint32_t type;
       uint32_t flags;
@@ -219,13 +214,6 @@ namespace Elf
       void *mem;
   };
 
-  template <typename _class, typename _data>
-    class segment_data : public virtual segment
-    {
-      public:
-        segment_data (void *, void *) throw ();
-    };
-
   template <typename _type>
     class segment_type : public virtual segment
     {
@@ -243,29 +231,6 @@ namespace Elf
         std::string interp;
     };
 
-  template <typename _class, typename _data, typename _type>
-    class segment_real : public segment_data<_class, _data>, public segment_type<_type>
-    {
-      public:
-        segment_real (void *, void *) throw ();
-    };
-
-  template <typename _class, typename _data>
-    class segment_real<_class, _data, segment_type_UNDEFINED>
-    : public segment_data<_class, _data>, public segment_type<segment_type_UNDEFINED>
-    {
-      public:
-        segment_real (void *a, void *b) throw () : segment_data<_class, _data> (a, b) { }
-    };
-
-  template <typename _class, typename _data>
-    class segment_real<_class, _data, segment_type_INTERP>
-    : public segment_data<_class, _data>, public segment_type<segment_type_INTERP>
-    {
-      public:
-        segment_real (void *, void *) throw (std::bad_alloc);
-    };
-
   class dynamic
   {
     public:
@@ -276,8 +241,6 @@ namespace Elf
       uint64_t get_ptr () const throw () { return ptr; }
       const std::string &get_val_string () const throw () { return val_string; }
 
-      virtual void update_string_table (file *, uint16_t) throw (std::bad_alloc) = 0;
-
     protected:
       int64_t tag;
       uint64_t val;
@@ -287,19 +250,11 @@ namespace Elf
       std::string val_string;
   };
 
-  template <typename _class, typename _data>
-    class dynamic_data : public dynamic
-    {
-      public:
-        dynamic_data (void *) throw ();
-
-        void update_string_table (file *, uint16_t) throw (std::bad_alloc);
-    };
-
   class symbol
   {
     public:
-      virtual ~symbol () throw () {}
+      symbol() throw () : verdef(NULL), verneed(NULL) {}
+      virtual ~symbol() throw () {}
 
       uint8_t get_info () const throw () { return info; }
       uint16_t get_shndx () const throw () { return shndx; }
@@ -308,8 +263,9 @@ namespace Elf
       uint8_t get_bind () const throw () { return bind; }
       uint8_t get_type () const throw () { return type; }
       const std::string &get_name_string () const throw () { return name_string; }
-
-      virtual void update_string_table (file *, uint16_t) throw (std::bad_alloc) = 0;
+      std::string get_version() const throw (std::bad_alloc);
+      uint16_t get_version_data() const throw () { return versym; }
+      std::string get_name_version() const throw (std::bad_alloc);
 
     protected:
       uint32_t name;
@@ -321,28 +277,56 @@ namespace Elf
       uint8_t type;
 
       std::string name_string;
+
+      const version_definition *verdef;
+      const version_requirement_entry *verneed;
+      uint16_t versym;
   };
 
-  template <typename _class, typename _data>
-    class symbol_data : public symbol
-    {
-      public:
-        symbol_data (void *) throw ();
+  class version_definition
+  {
+    public:
+      virtual ~version_definition() throw () { }
 
-        void update_string_table (file *, uint16_t) throw (std::bad_alloc);
-    };
+      uint16_t get_ndx() const throw () { return ndx; }
+      const std::vector<std::string> &get_names() const throw () { return names_string; }
 
-  template <typename from, typename type>
-    struct convert
-    { };
+    protected:
+      uint16_t ndx;
 
-  template <typename type>
-    struct convert<file_data_2LSB, type> : public endian::convert<endian::little_endian, type>
-    { };
+      std::vector<uint32_t> names;
 
-  template <typename type>
-    struct convert<file_data_2MSB, type> : public endian::convert<endian::big_endian, type>
-    { };
+      std::vector<std::string> names_string;
+  };
+
+  class version_requirement_entry;
+
+  class version_requirement
+  {
+    public:
+      virtual ~version_requirement () throw () { }
+
+      const std::vector<version_requirement_entry *> &get_entries () const throw () { return entries; }
+
+    protected:
+      std::vector<version_requirement_entry *> entries;
+  };
+
+  class version_requirement_entry
+  {
+    public:
+      virtual ~version_requirement_entry () throw () { }
+
+      uint16_t get_other () const throw () { return other; }
+      const std::string &get_name() const throw () { return name_string; }
+
+    protected:
+      uint16_t flags;
+      uint16_t other;
+      uint32_t name;
+
+      std::string name_string;
+  };
 }
 
 #endif
