@@ -119,33 +119,55 @@ def library_depends_gcc_libnames(obj):
                 ret.append('-l%s' % match.group('lib'))
     return ' '.join(ret)
 
+class Symbol(object):
+    def __init__(self, name, version):
+        self.name, self.version = name, version
+
+    def __str__(self):
+        return "%s@%s" % (self.name, self.version)
+
+class UndefinedSymbol(Symbol):
+    def __init__(self, name, weak, version):
+        super(UndefinedSymbol, self).__init__(name, version)
+        self.weak = weak
+
 # Return undefined symbols in an object as a set of tuples (name, weakness)
 def undefined_symbols(obj):
     if not os.access(obj, os.F_OK):
         raise "Cannot find lib" + obj
 
-    result = set()
+    result = []
     output = command("mklibs-readelf", "--print-symbols-undefined", obj)
     for line in output:
-        name, weak_string = line.split(' ', 2)
-        weak = False
-        if weak_string == 'True':
-            weak = True
-        debug(DEBUG_SPAM, "undefined_symbols adding: %s, %s" % (name, weak))
-        result.add((name, weak))
+        name, weak_string, version = line.split()[:3]
+        result.append(UndefinedSymbol(name, bool(eval(weak_string)), version))
     return result
 
+class ProvidedSymbol(Symbol):
+    def __init__(self, name, version, default_version):
+        super(ProvidedSymbol, self).__init__(name, version)
+        self.default_version = default_version
+
+    def base_names(self):
+        if self.default_version and self.version != "Base":
+            return ["%s@%s" % (self.name, self.version), "%s@Base" % self.name]
+        return ["%s@%s" % (self.name, self.version)]
+
+    def linker_name(self):
+        if self.default_version or self.version == "Base":
+            return self.name
+        return "%s@%s" % (self.name, self.version)
+
 # Return a set of symbols provided by a library
 def provided_symbols(obj):
     if not os.access(obj, os.F_OK):
         raise "Cannot find lib" + obj
 
-    result = set()
+    result = []
     output = command("mklibs-readelf", "--print-symbols-provided", obj)
     for line in output:
-        name, weak = line.split(' ', 2)
-        debug(DEBUG_SPAM, "provided_symbols adding %s, %s" %(name, weak))
-        result.add(name)
+        name, weak, version, default_version_string = line.split()[:4]
+        result.append(ProvidedSymbol(name, version, bool(eval(default_version_string))))
     return result
     
 # Return real target of a symlink
@@ -203,6 +225,7 @@ def usage(was_err):
     print >> outfd, "  -d, --dest-dir DIRECTORY     create libraries in DIRECTORY"
     print >> outfd, "  -D, --no-default-lib         omit default libpath (", ':'.join(default_lib_path), ")"
     print >> outfd, "  -L DIRECTORY[:DIRECTORY]...  add DIRECTORY(s) to the library search path"
+    print >> outfd, "  -l LIBRARY                   add LIBRARY always"
     print >> outfd, "      --ldlib LDLIB            use LDLIB for the dynamic linker"
     print >> outfd, "      --libc-extras-dir DIRECTORY  look for libc extra files in DIRECTORY"
     print >> outfd, "      --target TARGET          prepend TARGET- to the gcc and binutils calls"
@@ -239,7 +262,7 @@ vers="0.12"
 os.environ['LC_ALL'] = "C"
 
 # Argument parsing
-opts = "L:DnvVhd:r:"
+opts = "L:DnvVhd:r:l:"
 longopts = ["no-default-lib", "dry-run", "verbose", "version", "help",
             "dest-dir=", "ldlib=", "libc-extras-dir=", "target=", "root="]
 
@@ -253,6 +276,7 @@ default_lib_path = ["/lib/", "/usr/lib/", "/usr/X11R6/lib/"]
 libc_extras_dir = "/usr/lib/libc_pic"
 target = ""
 root = ""
+force_libs = []
 so_pattern = re.compile("((lib|ld).*)\.so(\..+)*")
 script_pattern = re.compile("^#!\s*/")
 
@@ -280,6 +304,8 @@ for opt, arg in optlist:
         target = arg + "-"
     elif opt in ("-r", "--root"):
         root = arg
+    elif opt in ("-l",):
+        force_libs.append(arg)
     elif opt in ("--help", "-h"):
 	usage(0)
         sys.exit(0)
@@ -364,28 +390,33 @@ while 1:
         debug(DEBUG_VERBOSE, "Object:", obj)
 
     # calculate what symbols and libraries are needed
-    needed_symbols = set()              # set of (name, weakness-flag)
-    libraries = set()
+    needed_symbols = {}
+    libraries = set(force_libs)
     for obj in objects.values():
-        needed_symbols.update(undefined_symbols(obj))
+        for symbol in undefined_symbols(obj):
+            debug(DEBUG_SPAM, "needed_symbols adding %s, weak: %s" % (symbol, symbol.weak))
+            needed_symbols[str(symbol)] = symbol
         libraries.update(library_depends(obj))
 
     # calculate what symbols are present in small_libs and available_libs
-    present_symbols = set()
+    present_symbols = {}
     checked_libs = small_libs
     checked_libs.extend(available_libs)
     checked_libs.append(ldlib)
     for lib in checked_libs:
-        present_symbols.update(provided_symbols(lib))
+        for symbol in provided_symbols(lib):
+            debug(DEBUG_SPAM, "present_symbols adding %s" % symbol)
+            names = symbol.base_names()
+            for name in names:
+                present_symbols[name] = symbol
 
     # are we finished?
     num_unresolved = 0
-    present_symbols_elems = present_symbols
     unresolved = set()
-    for (symbol, is_weak) in needed_symbols:
-        if not symbol in present_symbols_elems:
-            debug(DEBUG_SPAM, "Still need:", symbol, `is_weak`)
-            unresolved.add((symbol, is_weak))
+    for name in needed_symbols:
+        if not name in present_symbols:
+            debug(DEBUG_SPAM, "Still need: %s" % name)
+            unresolved.add(name)
             num_unresolved = num_unresolved + 1
 
     debug (DEBUG_NORMAL, `len(needed_symbols)`, "symbols,",
@@ -396,9 +427,9 @@ while 1:
 
     if unresolved == previous_pass_unresolved:
         # No progress in last pass. Verify all remaining symbols are weak.
-        for (symbol, is_weak) in unresolved:
-            if not is_weak:
-                raise "Unresolvable symbol " + symbol
+        for name in unresolved:
+            if not needed_symbols[name].weak:
+                raise "Unresolvable symbol %s" % name
         break
 
     previous_pass_unresolved = unresolved
@@ -419,34 +450,34 @@ while 1:
             sys.exit("Library not found: " + library + " in path: "
                     + ':'.join(lib_path))
         symbols = provided_symbols(path)
-        library_symbols[library] = set()
-        library_symbols_used[library] = set()    
+        library_symbols[library] = {}
+        library_symbols_used[library] = set()
         for symbol in symbols:
-            if symbol_provider.has_key(symbol):
-                # in doubt, prefer symbols from libc
-                if re.match("^libc[\.-]", library):
-                    library_symbols[library].add(symbol)
-                    symbol_provider[symbol] = library
+            for name in symbol.base_names():
+                if name in symbol_provider:
+                    # in doubt, prefer symbols from libc
+                    if re.match("^libc[\.-]", library):
+                        library_symbols[library][name] = symbol
+                        symbol_provider[name] = library
+                    else:
+                        debug(DEBUG_SPAM, "duplicate symbol %s in %s and %s" % (symbol, symbol_provider[name], library))
                 else:
-                    debug(DEBUG_SPAM, "duplicate symbol", symbol, "in", 
-                          symbol_provider[symbol], "and", library)
-            else:
-                library_symbols[library].add(symbol)
-                symbol_provider[symbol] = library
+                    library_symbols[library][name] = symbol
+                    symbol_provider[name] = library
 
     # which symbols are actually used from each lib
-    for (symbol, is_weak) in needed_symbols:
-        if not symbol_provider.has_key(symbol):
-            if not is_weak:
-                raise "No library provides non-weak " + symbol
+    for name in needed_symbols:
+        if not name in symbol_provider:
+            if not needed_symbols[name].weak:
+                raise "No library provides non-weak %s" % symbol
         else:
-            lib = symbol_provider[symbol]
-            library_symbols_used[lib].add(symbol)
+            lib = symbol_provider[name]
+            library_symbols_used[lib].add(library_symbols[lib][name])
 
     # reduce libraries
     for library in libraries:
         debug(DEBUG_VERBOSE, "reducing", library)
-        debug(DEBUG_SPAM, "using: " + string.join(library_symbols_used[library]))
+        debug(DEBUG_SPAM, "using: " + ' '.join([str(i) for i in library_symbols_used[library]]))
         so_file = find_lib(library)
         if root and (re.compile("^" + root).search(so_file)):
             debug(DEBUG_VERBOSE, "no action required for " + so_file)
@@ -485,7 +516,7 @@ while 1:
                 # may segfault in ptmalloc_init due to undefined weak reference
                 extra_pre_obj.append(libc_extras_dir + "/soinit.o")
                 extra_post_obj.append(libc_extras_dir + "/sofini.o")
-                symbols.add('__dso_handle')
+                symbols.add(ProvidedSymbol('__dso_handle', 'Base', True))
 
             map_file = find_pic_map(library)
             if map_file:
@@ -494,7 +525,7 @@ while 1:
             # compile in only used symbols
             cmd = []
             cmd.append("-nostdlib -nostartfiles -shared -Wl,-soname=" + soname)
-            cmd.extend(["-u%s" % a for a in symbols])
+            cmd.extend(["-u%s" % a.linker_name() for a in symbols])
             cmd.extend(["-o", dest_path + "/" + so_file_name + "-so"])
             cmd.extend(extra_pre_obj)
             cmd.append(pic_file)
